/**
# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# =============================================================================
*/
#ifndef INCLUDED_ORANG_ORANG_H
#define INCLUDED_ORANG_ORANG_H

#include <base.h>
#include <exception.h>
#include <table.h>

#include <graph.h>
#include <treedecomp.h>
#include <varorder.h>

#include <combine.h>
#include <marginalizer.h>
#include <operations/min.h>
#include <operations/count.h>
#include <operations/logsumprod.h>
#include <operations/dummy.h>

#include <task.h>

#include <merger.h>

#include <buckettree.h>

#endif
